from typing import *  # type: ignore

from ..common import URL
from .fundamental_component import FundamentalComponent

__all__ = [
    "Website",
]


class Website(FundamentalComponent):
    """
    # Website

    Displays a website.

    `Website` takes a URL as input and displays that website in your app.

    ## Attributes:

        `url`: The URL of the website you want to display.

    ## Example:

    A minimal example of a `Website` will be shown:

    ```python
    rio.Website(url=rio.URL("https://www.example.com"))
    ```
    """

    url: URL


Website._unique_id = "Website-builtin"



from typing import *  # type: ignore

import rio

from .. import components as comps


class BiographyPage(rio.Component):
    def build(self) -> rio.Component:
        thm = self.session.theme

        return rio.Column(
            rio.Rectangle(
                fill=rio.LinearGradientFill(
                    (thm.primary_palette.background, 0),
                    (thm.primary_palette.background.brighter(0.2), 1),
                ),
                height=1.0,
            ),
            rio.Column(
                comps.AboutMe(),
                rio.Grid(
                    (
                        comps.Contact(
                            "123 Main Street, Anytown, USA 12345",
                            "location-on",
                        ),
                        comps.Contact(
                            "Gitlab",
                            rio.URL(
                                "https://about.gitlab.com/images/press/logo/svg/gitlab-logo-500.svg"
                            ),
                            link=rio.URL("https://gitlab.com"),
                        ),
                    ),
                    (
                        comps.Contact(
                            "(123) 456-7890",
                            "call:fill",
                        ),
                        comps.Contact(
                            "GitHub",
                            rio.URL(
                                "https://github.githubassets.com/images/modules/logos_page/GitHub-Mark.png"
                            ),
                            link=rio.URL("https://github.com"),
                        ),
                    ),
                    (
                        comps.Contact(
                            "janedoe@email.com",
                            "mail",
                            link=rio.URL("mailto:janedoe@email.com"),
                        ),
                        comps.Contact(
                            "LinkedIn",
                            rio.URL(
                                "https://upload.wikimedia.org/wikipedia/commons/thumb/c/ca/LinkedIn_logo_initials.png/768px-LinkedIn_logo_initials.png"
                            ),
                            link=rio.URL("https://linkedin.com"),
                        ),
                    ),
                    row_spacing=1.5,
                    column_spacing=5,
                    width="grow",
                    margin_y=2,
                ),
                rio.Row(
                    comps.SkillBars(
                        {
                            "Python": 5,
                            "Rust": 3,
                            "SQL": 4,
                            "R": 3,
                        },
                        width="grow",
                        align_y=0,
                    ),
                    comps.SkillBars(
                        {
                            "Pandas": 5,
                            "Numpy": 5,
                            "SciPy": 3,
                            "Matplotlib": 4,
                            "Seaborn": 4,
                        },
                        width="grow",
                        align_y=0,
                    ),
                    comps.SkillBars(
                        {
                            "English": 5,
                            "Spanish": 3,
                            "French": 2,
                        },
                        width="grow",
                        align_y=0,
                    ),
                    width="grow",
                    spacing=5,
                ),
                comps.History(margin_x=10),
                rio.Text(
                    "Personal Projects",
                    style="heading2",
                    margin_x=8,
                    margin_top=3,
                ),
                comps.Projects(),
                rio.Overlay(
                    rio.IconButton(
                        icon="material/mail:fill",
                        margin_right=3,
                        margin_bottom=3,
                        align_x=1,
                        align_y=1,
                    )
                ),
                margin_y=4,
                spacing=2,
                width=40,
                align_x=0.5,
            ),
            align_y=0,
        )



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Card",
]


class Card(FundamentalComponent):
    """
    # Card

    A container that visually encompasses its child components.

    Cards are used to group related components together, and to visually
    separate them from other components, allowing you to display them in a
    structured way.

    Cards are also often used as large buttons. They can be configured to
    elevate slightly when the mouse hovers over them, indicating to the user
    that they support interaction.


    ## Attributes:

    `content`: The component to display inside the card.

    `corner_radius`: The radius of the card's corners. If set to `None`, it
            is picked from the active theme.

    `on_press`: An event handler that is called when the card is clicked.
            Note that attaching an even handler will also modify the appearance
            of the card, to signal the possible interaction to the user. See
            `elevate_on_hover` and `colorize_on_hover` for details.

    `elevate_on_hover`: Whether the card should elevate slightly when the
            mouse hovers over it. If set to `None` the card will elevate if
            an `on_press` event handler is attached.

    `colorize_on_hover`: Whether the card should change its color when the
            mouse hovers over it. If set to `None` the card will change its
            color if an `on_press` event handler is attached.

    `color`: The color scheme to use for the card. The color scheme controls
            the background color of the card, and the color of the text and
            icons inside it. Check `rio.Color` for details.


    ## Example:

    This minimal example will simply display a `Card` with the castle icon
    in it:

    ```python
    rio.Card(content=rio.Icon("material/castle"))
    ```

    `Card`s are commonly used to display content. You can easily make your Card
    interactive by adding a lambda function call to on_press:

    ```python
    class MyComponent(rio.Component):
        def build(self) -> rio.Component:
            # Create a row with an icon and text
            # as the card content
            card_content = rio.Row(
                rio.Icon(icon="material/castle"),
                rio.Text("Click me!"),
                spacing=1,
                # align the card content in the center
                align_x=0.5,
            )

            return rio.Card(
                content=card_content,
                on_press=lambda: print("Card clicked!"),
                elevate_on_hover=True,
            )
    ```


    You can also use a method for updating the input text and do whatever you
    want. Note that methods are handy if you want to do more than just updating
    the input text. For example run async code or update other components based
    on the input text:

    ```python
    class MyComponent(rio.Component):
        def on_press_card(self) -> None:
            # Do whatever you want when the card is pressed
            print("Card was pressed!")

        def build(self) -> rio.Component:
            # Create a row with an icon and text as the card content
            card_content = rio.Row(
                rio.Icon(icon="material/castle"),
                rio.Text("Click me!"),
                spacing=1,
                # Align the card content in the center
                align_x=0.5,
            )
            return rio.Card(
                content=card_content,
                on_press=self.on_press_card,
                elevate_on_hover=True,
            )
    ```
    """

    content: rio.Component
    _: KW_ONLY
    corner_radius: float | tuple[float, float, float, float] | None = None
    on_press: rio.EventHandler[[]] = None
    elevate_on_hover: bool | None = None
    colorize_on_hover: bool | None = None
    color: rio.ColorSet = "neutral"

    async def _on_message(self, msg: Any) -> None:
        # Trigger the press event
        await self.call_event_handler(self.on_press)

        # Refresh the session
        await self.session._refresh()

    def _custom_serialize(self) -> JsonDoc:
        thm = self.session.theme
        color = thm._serialize_colorset(self.color)

        report_press = self.on_press is not None

        return {
            "corner_radius": (
                thm.corner_radius_medium
                if self.corner_radius is None
                else self.corner_radius
            ),
            "reportPress": report_press,
            "elevate_on_hover": (
                report_press if self.elevate_on_hover is None else self.elevate_on_hover
            ),
            "colorize_on_hover": (
                report_press
                if self.colorize_on_hover is None
                else self.colorize_on_hover
            ),
            "color": color,
        }


Card._unique_id = "Card-builtin"



import hashlib
import os
import re
import secrets
import socket
from dataclasses import dataclass
from io import BytesIO, StringIO
from pathlib import Path
from typing import *  # type: ignore

import imy.assets
import imy.package_metadata
from PIL.Image import Image
from typing_extensions import Annotated
from yarl import URL

__all__ = [
    "EventHandler",
    "FileInfo",
    "ImageLike",
    "escape_markdown",
    "escape_markdown_code",
]


_SECURE_HASH_SEED: bytes = secrets.token_bytes(32)

# Expose common paths on the filesystem
PACKAGE_ROOT_DIR = Path(__file__).resolve().parent
PROJECT_ROOT_DIR = PACKAGE_ROOT_DIR.parent

GENERATED_DIR = PACKAGE_ROOT_DIR / "generated"

RIO_ASSETS_DIR = PACKAGE_ROOT_DIR / "assets"
HOSTED_ASSETS_DIR = RIO_ASSETS_DIR / "hosted"

RIO_LOGO_ASSET_PATH = HOSTED_ASSETS_DIR / "rio-logos/rio-logo-square.png"

SNIPPETS_DIR = PACKAGE_ROOT_DIR / "snippets" / "snippet-files"

if os.name == "nt":
    USER_CACHE_DIR = Path.home() / "AppData" / "Local" / "Cache"
    USER_CONFIG_DIR = Path.home() / "AppData" / "Roaming"
else:
    USER_CACHE_DIR = Path.home() / ".cache"
    USER_CONFIG_DIR = Path.home() / ".config"


# Constants & types
_READONLY = object()
T = TypeVar("T")
Readonly = Annotated[T, _READONLY]

ImageLike = Path | Image | URL | bytes


ASSET_MANGER: imy.assets.AssetManager = imy.assets.AssetManager(
    xz_dir=RIO_ASSETS_DIR,
    cache_dir=USER_CACHE_DIR / "rio",
    version=imy.package_metadata.get_package_version("rio-ui"),
)


# Precompiled regexes
MARKDOWN_ESCAPE = re.compile(r"([\\`\*_\{\}\[\]\(\)#\+\-.!])")
MARKDOWN_CODE_ESCAPE = re.compile(r"([\\`])")


def secure_string_hash(*values: str, hash_length: int = 32) -> str:
    """
    Returns a reproducible, secure hash for the given values.

    The order of values matters. In addition to the given values a global seed
    is added. This seed is generated once when the module is loaded, meaning the
    result is not suitable to be persisted across sessions.
    """

    hasher = hashlib.blake2b(
        _SECURE_HASH_SEED,
        digest_size=hash_length,
    )

    for value in values:
        hasher.update(value.encode("utf-8"))

    return hasher.hexdigest()


@dataclass(frozen=True)
class FileInfo:
    """
    Contains information about a file.

    When asking the user to select a file, this class is used to represent the
    file. It contains metadata about the file, and can also be used to access
    the file's contents.

    Be careful when running your app as a webserver, since files will need to be
    uploaded by the user, which is a potentially very slow operation.

    Attributes:
        name: The name of the file, including the extension.

        size_in_bytes: The size of the file, in bytes.

        media_type: The MIMe type of the file, for example `text/plain` or
            `image/png`.
    """

    name: str
    size_in_bytes: int
    media_type: str
    _contents: bytes

    async def read_bytes(self) -> bytes:
        """
        Asynchronously reads the entire file as `bytes`.

        Reads and returns the entire file as a `bytes` object. If you know that
        the file is text, consider using `read_text` instead.
        """
        return self._contents

    async def read_text(self, *, encoding: str = "utf-8") -> str:
        """
        Asynchronously reads the entire file as text.

        Reads and returns the entire file as a `str` object. The file is decoded
        using the given `encoding`. If you don't know that the file is valid
        text, use `read_bytes` instead.

        Args:
            encoding: The encoding to use when decoding the file.

        Raises:
            UnicodeDecodeError: The file could not be decoded using the given
                `encoding`.
        """
        return self._contents.decode(encoding)

    @overload
    async def open(self, type: Literal["r"]) -> StringIO: ...

    @overload
    async def open(self, type: Literal["rb"]) -> BytesIO: ...

    async def open(self, type: Literal["r", "rb"] = "r") -> StringIO | BytesIO:
        """
        Asynchronously opens the file, as though it were a regular file on this
        device.

        Opens and returns the file as a file-like object. If 'r' is specified,
        the file is opened as text. If 'rb' is specified, the file is opened as
        bytes.

        Args:
            type: The mode to open the file in. 'r' for text, 'rb' for bytes.

        Returns:
            A file-like object containing the file's contents.
        """
        # Bytes
        if type == "rb":
            return BytesIO(await self.read_bytes())

        # UTF
        if type == "r":
            return StringIO(await self.read_text())

        # Invalid
        raise ValueError("Invalid type. Expected 'r' or 'rb'.")


T = TypeVar("T")
P = ParamSpec("P")

EventHandler = Callable[P, Any | Awaitable[Any]] | None


def make_url_relative(base: URL, other: URL) -> URL:
    """
    Returns `other` as a relative URL to `base`. Raises a `ValueError` if
    `other` is not a child of `base`.

    This will never generate URLs containing `..`. If those would be needed a
    `ValueError` is raised.
    """
    # Verify the URLs have the same scheme and host
    if base.scheme != other.scheme:
        raise ValueError(
            f'URLs have different schemes: "{base.scheme}" and "{other.scheme}"'
        )

    if base.host != other.host:
        raise ValueError(f'URLs have different hosts: "{base.host}" and "{other.host}"')

    # Get the path segments of the URLs
    base_parts = base.parts
    other_parts = other.parts

    # Strip empty segments
    if base_parts and base_parts[-1] == "":
        base_parts = base_parts[:-1]

    if other_parts and other_parts[-1] == "":
        other_parts = other_parts[:-1]

    # See if the base is a parent of the other URL
    if base_parts != other_parts[: len(base_parts)]:
        raise ValueError(f'"{other}" is not a child of "{base}"')

    # Remove the common parts from the URL
    other_parts = other_parts[len(base_parts) :]
    return URL.build(
        path="/".join(other_parts),
        query=other.query,
        fragment=other.fragment,
    )


def escape_markdown(text: str) -> str:
    """
    Escape text such that it appears as-is when rendered as markdown.

    Given any text, this function returns a string which, when rendered as
    markdown, will look identical to the original text.
    """
    # TODO: Find a proper function for this. The current one is a total hack.
    return re.sub(MARKDOWN_ESCAPE, r"\\\1", text)


def escape_markdown_code(text: str) -> str:
    """
    Escape text such that it appears as-is inside a markdown code block.

    Given any text, this function returns a string which, when rendered inside
    a markdown code block, will look identical to the original text.
    """
    # TODO: Find a proper function for this. The current one is a total hack.
    return re.sub(MARKDOWN_CODE_ESCAPE, r"\\\1", text)


def choose_free_port(host: str) -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind((host, 0))
        return sock.getsockname()[1]


def port_is_free(host: str, port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        try:
            sock.bind((host, port))
            return True
        except OSError as err:
            return False


def ensure_valid_port(host: str, port: int | None) -> int:
    if port is None:
        return choose_free_port(host)

    return port


def first_non_null(*values: T | None) -> T:
    """
    Returns the first non-`None` value, or raises a `ValueError` if all values
    are `None`.
    """

    for value in values:
        if value is not None:
            return value

    raise ValueError("At least one value must be non-`None`")



import rio
import rio.cli


class ProjectPage(rio.Component):
    project: rio.cli.project.RioProject | None = None

    def __post_init__(self) -> None:
        self.project = rio.cli.project.RioProject.try_locate_and_load()

    async def _on_change_app_type(self, event: rio.DropdownChangeEvent) -> None:
        assert self.project is not None
        self.project.app_type = event.value
        self.project.write()
        await self.force_refresh()

    def build(self) -> rio.Component:
        # No project
        if self.project is None:
            return rio.Column(
                rio.Icon("material/error", width=4, height=4, fill="danger"),
                rio.Text(
                    "Couldn't find your project files. Do you have a `rio.toml` file?",
                    wrap=True,
                ),
            )

        # Project
        project_name = self.project.project_directory.name.strip().capitalize()

        # FIXME: The contents of the rio.toml file are currently completely
        # ignored (except for the project name). Display the data and edit the
        # file if the user changes the values!

        return rio.Column(
            rio.Text(
                project_name,
                style="heading2",
                align_x=0,
            ),
            rio.Text(
                str(self.project.project_directory),
                style="dim",
                margin_bottom=2,
                align_x=0,
            ),
            rio.Text(
                "To launch your project, Rio needs to know the name of your python module and in which variable you've stored your app. You can configure those here.",
                wrap=True,
            ),
            rio.TextInput(
                label="Main Module",
            ),
            rio.TextInput(
                label="App Variable",
            ),
            rio.Text(
                "Rio can create both apps and websites. Apps will launch in a separate window, while websites will launch in your browser. Which type is your project?",
                wrap=True,
            ),
            rio.Dropdown(
                label="Type",
                options={
                    "App": "app",
                    "Website": "website",
                },
            ),
            margin=1,
            align_y=0,
            spacing=0.5,
        )



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import Any

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "ColorPicker",
    "ColorChangeEvent",
]


@dataclass
class ColorChangeEvent:
    color: rio.Color


class ColorPicker(FundamentalComponent):
    """
    # ColorPicker

    Allows the user to pick an RGB(A) color.

    `ColorPicker` is a component that allows the user to pick a color. It
    displays a combination of colorful areas and sliders that the user can
    interact with to pick a color, and optionally an opacity slider to pick
    opacity.


    ## Attributes:

    `color`: The color that the user has picked.

    `pick_opacity`: Whether to allow the user to pick opacity. If `False`,
            the opacity slider will be hidden and the color value will be forced
            to be fully opaque.

    `on_change`: This event is triggered whenever the user changes the color.


    ## Example:

    This minimal example will simply display a `ColorPicker` with the default
    color red:

    ```python
    rio.ColorPicker(rio.Color.from_hex("#ff0000"))
    ```

    `ColorPicker`s are commonly used to let the user select a color. You can
    easily achieve this by using state bindings to save the color in a variable
    and using it in your application:

    ```python
    class MyComponent(rio.Component):
        color: rio.Color = rio.Color.from_hex("#ff0000")

        def build(self) -> rio.Component:
            return rio.ColorPicker(
                color=self.bind().color,
                pick_opacity=True,
            )
    ```

    If you want to make your `ColorPicker` more interactive, you can easily
    achieve this by adding a lambda function call to on_change:

    ```python
    class MyComponent(rio.Component):
        color: rio.Color = rio.Color.from_hex("#ff0000")

        def build(self) -> rio.Component:
            return rio.ColorPicker(
                color=self.bind().color,
                pick_opacity=True,
                on_change=lambda event: print(f"selected color: {event.color}"),
            )
    ```

    You can use an event handler to react to changes in the color. This example
    will print the color to the console whenever the user changes it:

    ```python
    class MyComponent(rio.Component):
        color: rio.Color = rio.Color.from_rgb(0.5, 0.5, 0.5)

        def on_change_color(self, event: rio.ColorChangeEvent) -> None:
            # You can do whatever you want in this method
            self.color = event.color
            print(f"Selected color: {event.color}")

        def build(self) -> rio.Component:
            return rio.Card(
                rio.ColorPicker(
                    color=self.color,
                    pick_opacity=True,
                    on_change=self.on_change_color,
                ),
            )
    ```
    """

    color: rio.Color
    _: KW_ONLY
    pick_opacity: bool = False
    on_change: rio.EventHandler[ColorChangeEvent] = None

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg

        color = rio.Color.from_rgb(*msg["color"])

        # Update the color
        self._apply_delta_state_from_frontend({"color": color})

        # Trigger the change event
        await self.call_event_handler(
            self.on_change,
            ColorChangeEvent(color),
        )

        # Refresh the session
        await self.session._refresh()

    def _custom_serialize(self) -> JsonDoc:
        return {
            "color": self.color.rgba,
        }


ColorPicker._unique_id = "ColorPicker-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

import rio

from . import component

__all__ = [
    "Banner",
]


class Banner(component.Component):
    r"""
    # Banner

    Displays a short message to the user.

    Banners can either show a short text message to the users, or disappear
    entirely if no message is set. Use them to inform the user about the result
    of an action, to give feedback on their input, or anything else that needs
    to be communicated.

    The messages have one of four levels: success, info, warning, and error. The
    levels control the appearance of the notification bar, and allow you to
    quickly communicate the nature of the message to the user.


    ## Attributes:

    `text`: The text to display. If `None` or empty, the banner will disappear
            entirely.

    `style`: Controls the appearance of the banner. The style is one of
            `info`, `success`, `warning` and `danger`. Depending on the value
            the banner may change its colors and icon.

    `markup`: Whether the text should be interpreted as Markdown. If `True`, the
            text will be rendered as Markdown, otherwise it will be rendered as
            plain text.

    `multiline`: Whether long text may be wrapped over multiple lines.
            Multiline banners are also styled slightly differently to make the
            icon fit their larger size. Use `"\n"` to add a line break.


    ## Example:

    This minimal example will simply display a banner with the text "This is a
    banner":

    ```python
    rio.Banner(text="This is a banner", style="info")
    ```

    `Banner`s are commonly used to inform the users of about the result
    of an action. You can easily achieve this by adding a banner with its text
    set to the result of the action. For example, you could show a banner with
    the text "Button pressed!" when a button is pressed:

    ```python
    class MyComponent(rio.Component):
        banner_text: str = ""

        def on_press_button(self) -> None:
            self.banner_text = "Button pressed!"

        def build(self) -> rio.Component:
            return rio.Column(
                rio.Banner(
                    text=self.banner_text,
                    style="info",
                ),
                rio.Button(
                    content="Press me!",
                    on_press=self.on_press_button,
                ),
                spacing=1,
            )
    ```
    """

    text: str | None
    style: Literal["info", "success", "warning", "danger"]

    _: KW_ONLY
    markup: bool = False
    multiline: bool = False

    def build(self) -> rio.Component:
        # Early out: Nothing to show
        if self.text is None:
            return rio.Spacer(width=0, height=0)

        text = self.text.strip()
        if not text:
            return rio.Spacer(width=0, height=0)

        # Prepare the style
        if self.style == "info":
            style_name = "secondary"
            icon = "material/info"
        elif self.style == "success":
            style_name = "success"
            icon = "material/check-circle"
        elif self.style == "warning":
            style_name = "warning"
            icon = "material/warning"
        elif self.style == "danger":
            style_name = "danger"
            icon = "material/error"
        else:
            raise ValueError(f"Invalid style: {self.style}")

        # Prepare the text child
        if self.markup:
            text_child = rio.Markdown(
                text,
                width="grow",
            )
        else:
            text_child = rio.Text(
                text,
                width="grow",
                wrap=self.multiline,
            )

        # Build the result
        if self.multiline:
            return rio.Card(
                content=rio.Row(
                    rio.Icon(
                        icon,
                        width=2.5,
                        height=2.5,
                        align_y=0,
                    ),
                    text_child,
                    spacing=1.5,
                    margin=1.5,
                ),
                color=style_name,
                corner_radius=self.session.theme.corner_radius_medium,
                elevate_on_hover=False,
            )

        return rio.Card(
            content=rio.Row(
                rio.Icon(icon),
                text_child,
                spacing=0.8,
                margin=0.5,
                align_x=0.5,
            ),
            color=style_name,
            corner_radius=self.session.theme.corner_radius_small,
            elevate_on_hover=False,
        )



import urllib.request
from pathlib import Path

import pandas as pd
import plotly.graph_objs as go

import rio

# Load the dataset
CSV_PATH = Path(__file__).parent / "smartphone-sales.csv"
CSV_URL = "https://TODO-real-url.com"

if not CSV_PATH.exists():
    with urllib.request.urlopen(CSV_URL) as response:
        CSV_PATH.write_bytes(response.read())

raw_df = pd.read_csv(CSV_PATH)
raw_df = raw_df.head(1000)


# <component>
class InteractivePlot(rio.Component):
    x_axis: str = raw_df.columns[0]
    y_axis: str = raw_df.columns[1]

    remove_outliers: bool = False

    async def on_download_csv(self) -> None:
        # Build a CSV file from the selected columns
        selected_df = raw_df[[self.x_axis, self.y_axis]]
        csv = selected_df.to_csv(index=False)

        # Save the file
        await self.session.save_file(
            file_name="scatter.csv",
            file_contents=csv,
        )

    def filter_data(self) -> pd.DataFrame:
        """
        Remove outliers from the dataset, by keeping only the tenth to ninetieth
        percentile.
        """
        result = raw_df

        # Remove outliers in the x-axis
        series = result[self.x_axis]

        if series.dtype in [int, float]:
            lower = series.quantile(0.10)
            upper = series.quantile(0.90)

            result = result[(series > lower) & (series < upper)]

        # Remove outliers in the y-axis
        series = result[self.y_axis]

        if series.dtype in [int, float]:
            lower = series.quantile(0.10)
            upper = series.quantile(0.90)

            result = result[(series > lower) & (series < upper)]

        return result

    def build(self) -> rio.Component:
        # Fetch the data
        if self.remove_outliers:
            df = self.filter_data()
        else:
            df = raw_df

        # Build the plot
        return rio.Row(
            rio.Plot(
                go.Figure(
                    go.Scatter(
                        x=df[self.x_axis],
                        y=df[self.y_axis],
                        mode="markers",
                    ),
                ),
                width="grow",
                height="grow",
                background=rio.Color.TRANSPARENT,
            ),
            rio.Card(
                rio.Column(
                    rio.Text(
                        "Data Sources",
                        style="heading3",
                    ),
                    rio.Dropdown(
                        label="X Axis",
                        options=list(df.columns),
                        selected_value=self.bind().x_axis,
                    ),
                    rio.Dropdown(
                        label="Y Axis",
                        options=list(df.columns),
                        selected_value=self.bind().y_axis,
                    ),
                    rio.Text(
                        "Filters",
                        style="heading3",
                    ),
                    rio.Row(
                        rio.Text("Remove Outliers"),
                        rio.Spacer(),
                        rio.Switch(
                            is_on=self.bind().remove_outliers,
                        ),
                    ),
                    rio.Spacer(),
                    rio.Button(
                        "Download CSV",
                        icon="material/download",
                        on_press=self.on_download_csv,
                    ),
                    spacing=0.8,
                    margin=1,
                ),
                margin=1,
            ),
        )


# </component>



from __future__ import annotations

from typing import *  # type: ignore

import rio

# <additional-imports>
from .. import conversation

# </additional-imports>


# <component>
class ChatMessage(rio.Component):
    """
    This component displays a single chat message, allowing you can create a
    message history by stacking multiple instances of this component vertically.
    """

    model: conversation.ChatMessage

    def build(self) -> rio.Component:
        # User messages look slightly different from the bot's responses. This
        # makes it easer for the user to distinguish between the two.
        if self.model.role == "user":
            icon = "rio/logo"
            color = "neutral"
        else:
            icon = "material/twinkle"
            color = "background"

        return rio.Row(
            # Display an icon on the left side of the message. It is wrapped
            # inside of a card to give it a circular shape.
            rio.Card(
                rio.Icon(
                    icon,
                    width=2,
                    height=2,
                    margin=0.8,
                ),
                # Using an enormous corner radius ensures that the card is fully
                # circular.
                corner_radius=99999,
                color="neutral",
                margin=0.5,
                align_y=0,
            ),
            # The message content is displayed inside of a second card. By using
            # markdown to show the message the user and chatbot can apply
            # formatting to their messages.
            rio.Card(
                rio.MarkdownView(
                    self.model.text,
                    margin=1.5,
                ),
                width="grow",
                color=color,
            ),
        )


# </component>



from __future__ import annotations

# <additional-imports>
import functools
from dataclasses import field
from typing import *  # type: ignore

import rio

# </additional-imports>


# <component>
class TicTacToePage(rio.Component):
    """
    This page contains the Tic Tac Toe game. It consists of 9 fields, which
    display either an X or an O. Players can press on empty fields to set their
    respective mark. You know the game.

    Since this game is so simple, there is no further components in the project.
    Everything is contained right in this class.
    """

    # The contents of all fields
    fields: list[Literal["X", "O", ""]] = field(default_factory=lambda: [""] * 9)

    # The player who is currently on turn
    player: Literal["X", "O"] = "X"

    # The winner of the game, if any
    winner: Literal["X", "O", "draw"] | None = None

    # If there is a winner, these are the indices of the fields which made them
    # win
    winning_indices: set[int] = field(default_factory=set)

    def find_winner(self) -> None:
        """
        Look if there is a winner on the board, and stores it in the component's
        state. Also updates the winning indices accordingly.
        """
        # Create a list of all possible winning combinations. If all of these
        # indices have the same value, then that value is the winner.
        winning_combinations = [
            [0, 1, 2],  # Top row
            [3, 4, 5],  # Middle row
            [6, 7, 8],  # Bottom row
            [0, 3, 6],  # Left column
            [1, 4, 7],  # Middle column
            [2, 5, 8],  # Right column
            [0, 4, 8],  # Top-left to bottom-right diagonal
            [2, 4, 6],  # Top-right to bottom-left diagonal
        ]

        # Look for winners
        for combination in winning_combinations:
            values = [self.fields[i] for i in combination]
            if values.count("X") == 3:
                self.winner = "X"
                self.winning_indices = set(combination)
                return

            if values.count("O") == 3:
                self.winner = "O"
                self.winning_indices = set(combination)
                return

        # If no fields are empty, it's a draw
        if "" not in self.fields:
            self.winner = "draw"
            return

        # There is no winner yet
        pass

    async def on_press(self, index: int) -> None:
        """
        This function reacts to presses on the fields, and updates the game
        state accordingly.
        """

        # If the game is already over, don't do anything
        if self.winner:
            return

        # Set the field on that index
        self.fields[index] = self.player

        # Next player
        self.player = "X" if self.player == "O" else "O"

        # See if there is a winner
        self.find_winner()

    def on_reset(self) -> None:
        """
        Reset the game to its initial state.
        """
        self.fields = [""] * 9
        self.player = "X"
        self.winner = None
        self.winning_indices = set()

    def build(self) -> rio.Component:
        # Spawn components for the fields
        field_components: list[rio.Component] = []
        field_size = 3

        for index, field in enumerate(self.fields):
            # Allow the player to press on empty fields. Since buttons would
            # look out of place here, cards are a nice alternative.
            if field == "":
                field_components.append(
                    rio.Card(
                        content=rio.Spacer(
                            width=field_size,
                            height=field_size,
                        ),
                        on_press=functools.partial(self.on_press, index),
                    )
                )

            # For fields that already contain an X or O, show the respective
            # icon
            else:
                # Vary the color based on the player
                color = rio.Color.RED if field == "X" else rio.Color.BLUE

                # If a player has won highlight the winning fields
                if self.winner in ("X", "O") and index not in self.winning_indices:
                    color = color.replace(opacity=0.3)

                # Create the icon to represent the field
                field_components.append(
                    rio.Icon(
                        "material/close" if field == "X" else "material/circle",
                        fill=color,
                        width=field_size,
                        height=field_size,
                    )
                )

        # Come up with a status message
        if self.winner in ("X", "O"):
            message = f"{self.winner} has won!"
        elif self.winner == "draw":
            message = "It's a draw!"
        else:
            message = f"{self.player}'s turn"

        # Arrange all components in a grid
        return rio.Column(
            rio.Text(message, style="heading1"),
            rio.Grid(
                field_components[0:3],
                field_components[3:6],
                field_components[6:9],
                row_spacing=1,
                column_spacing=1,
                align_x=0.5,
            ),
            rio.Button(
                "Reset",
                icon="material/refresh",
                style="plain",
                on_press=self.on_reset,
            ),
            spacing=2,
            margin=2,
            align_x=0.5,
            align_y=0.0,
        )


# </component>



from __future__ import annotations

from dataclasses import KW_ONLY
from datetime import timedelta
from typing import Literal

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Slideshow",
]


class Slideshow(FundamentalComponent):
    """
    # Slideshow

    Prominently switch between multiple components based on a timer.

    The `Slideshow` component is a container that can hold multiple components,
    and will display them one after the other, with smooth transitions in
    between. These are very useful for displaying a series of demos or news to
    visitors.


    ## Attributes:

    `children`: The components to transition between.

    `linger_time`: The time in seconds to display each component before
            switching to the next one.

    `corner_radius`: How rounded the slideshow's corners should be. If set to
            `None`, the slideshow will use a default corner radius from the
            current theme.


    ## Example:

    A minimal example of `Slideshow` displaying two images will be shown:

    ```python
    from pathlib import Path

    PATH = Path(__file__).parent

    rio.Slideshow(
        rio.Image(PATH / "first.jpg"),
        rio.Image(PATH / "second.jpg"),
        linger_time=5,
    )
    ```
    """

    children: list[rio.Component]
    _: KW_ONLY
    linger_time: float
    corner_radius: None | float | tuple[float, float, float, float]

    def __init__(
        self,
        *children: rio.Component,
        linger_time: float | timedelta = timedelta(seconds=10),
        corner_radius: None | float | tuple[float, float, float, float] = None,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        if isinstance(linger_time, timedelta):
            linger_time = linger_time.total_seconds()

        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.children = list(children)
        self.linger_time = linger_time
        self.corner_radius = corner_radius

    def _custom_serialize(self) -> JsonDoc:
        # Serialize the corner radius
        if self.corner_radius is None:
            thm = self.session.theme

            corner_radius = (
                thm.corner_radius_medium,
                thm.corner_radius_medium,
                thm.corner_radius_medium,
                thm.corner_radius_medium,
            )

        elif isinstance(self.corner_radius, (int, float)):
            corner_radius = (
                self.corner_radius,
                self.corner_radius,
                self.corner_radius,
                self.corner_radius,
            )

        else:
            corner_radius = self.corner_radius

        return {
            "corner_radius": corner_radius,
        }


Slideshow._unique_id = "Slideshow-builtin"



"""
Contains processing specific to the RIO project.
"""

from typing import *  # type: ignore

import rio

from . import models


def find_items_needing_documentation() -> Iterable[Type | Callable]:
    """
    Find all classes and functions in `Rio` that need to be documented.
    """

    # Hardcoded items
    yield rio.App
    yield rio.AssetError
    yield rio.Color
    yield rio.ColorChangeEvent
    yield rio.CursorStyle
    yield rio.DrawerOpenOrCloseEvent
    yield rio.DropdownChangeEvent
    yield rio.escape_markdown_code
    yield rio.escape_markdown
    yield rio.FileInfo
    yield rio.Font
    yield rio.KeyDownEvent
    yield rio.KeyPressEvent
    yield rio.KeyUpEvent
    yield rio.MouseDownEvent
    yield rio.MouseEnterEvent
    yield rio.MouseLeaveEvent
    yield rio.MouseMoveEvent
    yield rio.MouseUpEvent
    yield rio.NavigationFailed
    yield rio.NumberInputChangeEvent
    yield rio.NumberInputConfirmEvent
    yield rio.Page
    yield rio.PopupOpenOrCloseEvent
    yield rio.RevealerChangeEvent
    yield rio.Session
    yield rio.TextInputChangeEvent
    yield rio.TextInputConfirmEvent
    yield rio.TextStyle
    yield rio.Theme
    yield rio.UserSettings

    # Yield classes that also need their children documented
    to_do = [
        rio.Fill,
        rio.Component,
    ]

    while to_do:
        cur = to_do.pop()

        # Queue the children
        to_do.extend(cur.__subclasses__())

        # Hardcoded items that DON'T need documentation
        if cur.__name__ in (
            "ClassContainer",
            "FundamentalComponent",
            "RootContainer",
        ):
            continue

        # Skip anything not in the `rio` module
        if not cur.__module__.startswith("rio"):
            continue

        # Internal
        if not cur.__name__.startswith("_"):
            yield cur


def postprocess_class_docs(docs: models.ClassDocs) -> None:
    """
    Perform RIO specific post-processing on the component, such as stripping out
    internal attributes and functions.
    """

    # Strip default docstrings created by dataclasses
    if docs.summary is not None and docs.summary.startswith(f"{docs.name}("):
        docs.summary = None
        docs.details = None

    # Strip internal attributes
    index = 0
    while index < len(docs.attributes):
        attr = docs.attributes[index]

        # Decide whether to keep it
        keep = not attr.name.startswith("_")

        # Strip it out, if necessary
        if keep:
            index += 1
        else:
            del docs.attributes[index]

    # Skip internal functions
    index = 0
    while index < len(docs.functions):
        func = docs.functions[index]

        # Decide whether to keep it

        # Internal methods start with an underscore
        keep = not func.name.startswith("_")

        # Some methods in components are meant to be used by the user, but only
        # when they're the one creating the component. For example, the `build`
        # method is only intended to be used by the component itself, and
        # documenting it would be pointless at best, and confusing at worst.
        is_inherited_protected_method = docs.name != "Component" and func.name in (
            "build",
            "call_event_handler",
            "force_refresh",
        )
        keep = keep and not is_inherited_protected_method

        # Strip lambdas
        keep = keep and func.name != "<lambda>"

        # Make sure to keep the constructor
        keep = keep or func.name == "__init__"

        # Some classes are not meant to be constructed by the user. Strip their
        # constructor.
        if (
            docs.name
            in (
                "FileInfo",
                "Session",
            )
            and func.name == "__init__"
        ):
            keep = False

        # Strip it out, if necessary
        if keep:
            index += 1
        else:
            del docs.functions[index]

    # Additional per-function post-processing
    for func_docs in docs.functions:
        # Strip the ridiculous default docstring created by dataclasses
        #
        # FIXME: Not working for some reason
        if (
            func_docs.name == "__init__"
            and func_docs.summary
            == "Initialize self. See help(type(self)) for accurate signature."
        ):
            func_docs.summary = None
            func_docs.details = None

        # Inject a short description for `__init__` if there is none.
        if func_docs.name == "__init__" and func_docs.summary is None:
            func_docs.summary = f"Creates a new `{docs.name}` instance."

    # TODO: Strip out anything `Session` inherits from `unicall`


def postprocess_component_docs(docs: models.ClassDocs) -> None:
    # Apply the standard class post-processing
    postprocess_class_docs(docs)



from pathlib import Path
from typing import *  # type: ignore

import keyring
import platformdirs
import tomlkit
from revel import fatal

from . import rio_api


class CliInstance:
    """
    Singleton class which abstracts away the environment for the Rio CLI. It
    handles access to commonly used functionality such as the configuration file
    and the API client.
    """

    _instance = None
    _config: tomlkit.TOMLDocument
    _api_client: rio_api.RioApi | None

    def __new__(cls):
        # Already initialized?
        if cls._instance is not None:
            return cls._instance

        # Nope, create a new instance
        self = cls._instance = super(CliInstance, cls).__new__(cls)

        # Read the config
        try:
            config_dir = Path(platformdirs.user_config_dir("rio"))
            self._config = tomlkit.loads((config_dir / "config.toml").read_text())
        except FileNotFoundError:
            self._config = tomlkit.document()
        except OSError as err:
            fatal(f"Couldn't read Rio's configuration: {err}")

        # Api client
        self._api_client = None

        return self

    async def __aenter__(self) -> "CliInstance":
        return self

    async def __aexit__(self, *args) -> None:
        if self._api_client is not None:
            await self._api_client.close()

    @property
    def auth_token(self) -> str | None:
        """
        Tries to get the auth token from the keyring. Returns `None` if no token
        was stored in the keyring yet.
        """
        return keyring.get_password("rio", "rioApiAuthToken")

    @auth_token.setter
    def auth_token(self, value: str) -> None:
        """
        Stores the given auth token in the keyring.
        """
        keyring.set_password("rio", "rioApiAuthToken", value)

    async def get_api_client(self, *, logged_in: bool = True) -> rio_api.RioApi:
        """
        Return an API client for the Rio API. If `logged_in` is True, the
        client will be authenticated, prompting the user to login if necessary.
        """
        # If there is no client yet, create one
        if self._api_client is None:
            self._api_client = rio_api.RioApi()

        # Authenticate, if necessary
        if logged_in and not self._api_client.is_logged_in:
            raise NotImplementedError("TODO: Login")

        return self._api_client



from __future__ import annotations

from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Text",
]


class Text(FundamentalComponent):
    """
    # Text

    Displays unformatted text.

    `Text` displays text without any formatting, making it one of the most
    commonly used components in Rio.

    While the text itself is unformatted, you can still control the style of
    the text using the `style` attribute. This allows you to change the font
    size, color, and more.


    ## Attributes:

    `text`: The text to display.

    `multiline`: Whether the text may be split into multiple lines if not
        enough space is available.

    `selectable`: Whether the text can be selected by the user.

    `style`: The style of the text. This can either be a `TextStyle` instance,
        or one of the built-in styles: `heading1`, `heading2`, `heading3`,
        `text` or `dim`.


    ## Example:

    A minimal example of a `Text` will be shown:

    ```python
    rio.Text("Hello, world!")
    ```
    """

    text: str
    selectable: bool = True
    style: (
        Literal["heading1", "heading2", "heading3", "text", "dim"] | rio.TextStyle
    ) = "text"
    justify: Literal["left", "right", "center", "justify"] = "center"
    wrap: bool | Literal["ellipsize"] = False

    def _custom_serialize(self) -> JsonDoc:
        # Serialization doesn't handle unions. Hence the custom serialization
        # here
        if isinstance(self.style, str):
            style = self.style
        else:
            style = self.style._serialize(self.session)

        return {
            "style": style,
            "wrap": self.wrap,
        }

    def __repr__(self) -> str:
        if len(self.text) > 40:
            text = self.text[:40] + "..."
        else:
            text = self.text

        return f"<{type(self).__name__} id:{self._id} text:{text!r}>"


Text._unique_id = "Text-builtin"



import dataclasses
import functools
import re
from dataclasses import KW_ONLY
from typing import *  # type: ignore

import fuzzywuzzy.fuzz

import rio
import rio.icon_registry

from . import sample_icons_grid

T = TypeVar("T")


# Replace all sequences non-alphanumeric characters with a single dot
NORMALIZATION_PATTERN = re.compile(r"[^a-zA-Z0-9]+")

# Cache for all icons known to Rio. Finding the icons requires file system
# access and is slow - this caches them.
#
# The entires are (icon_set, icon_name, {variant_name}) tuples. The variant name
# is a set of all variants that exist for this icon.
ALL_AVAILABLE_ICONS: list[tuple[str, str, tuple[str | None, ...]]] | None = None


def normalize_for_search(text: str) -> str:
    """
    Normalize a string for search, making search more forgiving.
    """
    text = NORMALIZATION_PATTERN.sub(".", text)
    text = text.strip(".")
    text = text.lower()
    return text


def search(
    needle: str,
    haystack: list[T],
    *,
    min_results: int = 5,
    threshold: float = 0.85,
    key: Callable[[T], str] = lambda x: x,  # type: ignore
) -> list[tuple[T, float]]:
    """
    Given a set of candidate strings, return the best matches for the provided
    search string. The result is a list of (string, score) tuples, sorted by
    score. The score is 1.0 for a perfect match and 0.0 for no match.
    """
    # Convert the haystack elements to strings using the provided key function
    haystack_strs = [key(item) for item in haystack]

    # Use fuzzywuzzy to get the matches and their scores
    scores = [
        fuzzywuzzy.fuzz.partial_ratio(needle, haystack_strs)
        for haystack_strs in haystack_strs
    ]

    # Sort the original values by their score
    scored_haystack = sorted(
        zip(haystack, scores),
        key=lambda x: x[1],
        reverse=True,
    )

    # Apply a threshold, but never remove all results
    result = []
    for ii, (item, score) in enumerate(scored_haystack):
        score = score / 100

        if score < threshold and ii >= min_results:
            break

        result.append((item, score))

    return result


def get_available_icons() -> list[tuple[str, str, tuple[str | None, ...]]]:
    """
    Return a list of all available icons. The result is cached, so subsequent
    calls are fast.

    The result is a list of tuples (icon_set, icon_name, {variant_names}).
    """
    global ALL_AVAILABLE_ICONS

    # Cached?
    if ALL_AVAILABLE_ICONS is not None:
        return ALL_AVAILABLE_ICONS

    # Nope, get to work. Iterate over all icons and variants and group them.
    registry = rio.icon_registry.IconRegistry.get_singleton()
    result_dict: dict[str, tuple[str, str, set[str | None]]] = {}

    for icon_set in registry.all_icon_sets():
        for icon_name, variant_name in registry.all_icons_in_set(icon_set):
            key = f"{icon_set}/{icon_name}"

            try:
                _, _, variants = result_dict[key]
            except KeyError:
                variants = {variant_name}
                result_dict[key] = (icon_set, icon_name, variants)
            else:
                variants.add(variant_name)

    # Drop the dict
    as_list: list[Any] = list(result_dict.values())

    # Sort the result
    as_list.sort(key=lambda x: x[1])

    for ii, (icon_set, icon_name, variants) in enumerate(as_list):
        variants = tuple(sorted(variants, key=lambda x: "" if x is None else x))
        as_list[ii] = (icon_set, icon_name, variants)

    # Cache and return
    ALL_AVAILABLE_ICONS = as_list  # type: ignore
    return ALL_AVAILABLE_ICONS  # type: ignore


class IconsPage(rio.Component):
    _: KW_ONLY
    search_text: str = ""
    matches: list[tuple[str, str, tuple[str | None, ...]]] = dataclasses.field(
        default_factory=list
    )

    selected_icon: str | None = None  # The selected icon, **including the icon set**
    selected_icon_available_variants: tuple[str | None, ...] = dataclasses.field(
        default_factory=tuple
    )
    selected_variant: str | None = None
    selected_fill: Literal[
        "primary",
        "secondary",
        "success",
        "warning",
        "danger",
        "keep",
        "dim",
    ] = "keep"

    def _on_search_text_change(self, *_) -> None:
        # No search -> no matches
        self.matches = []
        normalized_search_text = self.search_text.strip()

        if not normalized_search_text:
            return

        # TODO: Add a proper search
        scored_matches = search(
            normalized_search_text,
            get_available_icons(),
            key=lambda x: x[1],
        )

        self.matches = [x for x, _ in scored_matches]

    async def _on_select_icon_by_name(self, icon_name: str) -> None:
        # Parse the icon name
        icon_set, icon_name, icon_variant = (
            rio.icon_registry.IconRegistry.parse_icon_name(icon_name)
        )

        # Find all available variants
        for (
            other_icon_set,
            other_icon_name,
            available_variants,
        ) in get_available_icons():
            if icon_name == other_icon_name and icon_set == other_icon_set:
                break
        else:
            assert (
                False
            ), f"There is no icon named `{icon_name}` in the `{icon_set}` icon set"

        # Delegate to the regular function for this
        self._on_select_icon(icon_set, icon_name, available_variants)

        # Set any state that wasn't handled by the previous call
        self.search_text = icon_name
        self.selected_variant = icon_variant

        self._on_search_text_change()

    def _on_select_icon(
        self,
        icon_set: str,
        icon_name: str,
        available_variants: tuple[str | None, ...],
    ) -> None:
        self.selected_icon = icon_set + "/" + icon_name
        self.selected_variant = available_variants[0]
        self.selected_icon_available_variants = available_variants

    def _on_select_variant(self, variant: str | None) -> None:
        self.selected_variant = variant

    def build_details(self) -> rio.Component:
        assert self.selected_icon is not None
        children = []

        # Heading
        children.append(rio.Text("Configure", style="heading3", align_x=0))

        # Fill
        children.append(
            rio.Dropdown(
                label="Fill",
                options=[
                    "primary",
                    "secondary",
                    "success",
                    "warning",
                    "danger",
                    "keep",
                    "dim",
                ],
                selected_value=self.bind().selected_fill,
            )
        )

        # If there is variations of this icon allow the user to select one
        # if len(self.selected_icon_available_variants) > 1:
        if True:
            variant_buttons = []

            for variant in self.selected_icon_available_variants:
                full_name = (
                    f"{self.selected_icon}:{variant}" if variant else self.selected_icon
                )

                variant_buttons.append(
                    rio.Container(
                        rio.IconButton(
                            full_name,
                            style=(
                                "minor" if variant == self.selected_variant else "plain"
                            ),
                            on_press=functools.partial(
                                self._on_select_variant, variant
                            ),
                        ),
                        width="grow",
                        align_y=0.5,
                    )
                )

            children.append(rio.Row(*variant_buttons, width="grow"))

        # Which parameters should be passed?
        params_dict = {
            "width": "2.5",
            "height": "2.5",
        }

        if self.selected_fill != "keep":
            params_dict["fill"] = repr(self.selected_fill)

        # Prepare the source code
        selected_icon_full_name = self.selected_icon + (
            "" if self.selected_variant is None else ":" + self.selected_variant
        )

        params_list = [repr(selected_icon_full_name)] + [
            f"{key}={value}" for key, value in params_dict.items()
        ]

        params_str = ",\n    ".join(params_list)
        python_source = f"rio.Icon(\n    {params_str},\n)"

        # Code sample
        children.append(
            rio.Text(
                "Example Code",
                style="heading3",
                align_x=0,
                margin_top=1,
            )
        )

        children.append(
            rio.Markdown(
                f"""
Use the `rio.Icon` component like this:

```python
{python_source}
```
"""
            )
        )

        # Combine everything
        return rio.Column(*children, spacing=0.8)

    def build(self) -> rio.Component:
        children: list[rio.Component] = []
        has_search = bool(self.search_text)

        # Top Decoration
        if not has_search:
            children.append(
                rio.Text(
                    "Rio comes with a large array of icons out of the box. You can find them all here.",
                    wrap=True,
                    style=rio.TextStyle(
                        font_size=1.1,
                        font_weight="bold",
                        fill=self.session.theme.secondary_color,
                    ),
                    margin_bottom=1,
                )
            )

        # Search field
        children.append(
            rio.TextInput(
                label="Search for an icon",
                text=self.bind().search_text,
                on_change=self._on_search_text_change,
            ),
        )

        # Display the top results in a grid
        if self.matches:
            results = []

            for icon_set, icon_name, icon_variants in self.matches[:50]:
                is_selected = self.selected_icon == f"{icon_set}/{icon_name}"

                results.append(
                    rio.IconButton(
                        f"{icon_set}/{icon_name}",
                        style="minor" if is_selected else "plain",
                        on_press=functools.partial(
                            self._on_select_icon, icon_set, icon_name, icon_variants
                        ),
                        key=f"{icon_set}/{icon_name}",
                    ),
                )

            children.append(
                rio.FlowContainer(
                    *results,
                    row_spacing=0.5,
                    column_spacing=0.5,
                    justify="left",
                    height="grow",
                    align_y=0,
                )
            )

        # Bottom Decoration
        if not has_search:
            children.append(
                sample_icons_grid.SampleIconsGrid(
                    height="grow",
                    margin_top=2,
                    on_select_icon=self._on_select_icon_by_name,
                )
            )

        # No results
        if has_search and not self.matches:
            children.append(
                rio.Container(
                    rio.Column(
                        rio.Icon(
                            "material/search",
                            width=4,
                            height=4,
                            fill="dim",
                            align_x=0.5,
                        ),
                        rio.Text(
                            "No results",
                            style="dim",
                        ),
                        spacing=1,
                        align_y=0.5,
                    ),
                    height="grow",
                )
            )

        # If an icon is selected, show its details
        if has_search and self.selected_icon is not None:
            children.append(self.build_details())

        # Combine everything
        return rio.ScrollContainer(
            rio.Column(
                *children,
                spacing=1,
                margin=1,
            ),
            scroll_x="never",
        )



from __future__ import annotations

import abc
import hashlib
import io
import os
import secrets
from pathlib import Path
from typing import *  # type: ignore

import httpx
from PIL.Image import Image
from yarl import URL

import rio

from .common import ImageLike
from .self_serializing import SelfSerializing

# Random bytes added during hashing to make any internal values extremely
# difficult to guess
_HASH_SALT = secrets.token_bytes(32)


def _securely_hash_bytes_changes_between_runs(data: bytes) -> bytes:
    """
    Returns an undefined, cryptographically secure hash of the given bytes. A
    random value is added to the hash to make it difficult to guess the
    original value. This random value is the same for all hashes generated by
    this function during the same run of the program, but changes between runs.
    """

    hasher = hashlib.sha256()
    hasher.update(_HASH_SALT)
    hasher.update(data)
    return hasher.digest()


_ASSETS: dict[tuple[bytes | Path | URL, str | None], Asset] = {}


class Asset(SelfSerializing):
    """
    Base class for assets - i.e. files that the client needs to be able to
    access. Assets can be hosted locally or remotely.

    Assets are "singletons", i.e. if you create two assets with the same input,
    they will be the same object:

        >>> Asset.new(Path("foo.png")) is Asset.new(Path("foo.png"))
        True

    To use an asset in a component, simply store it in the component's state. The
    asset will automatically register itself with the AppServer (if necessary)
    and serialize itself as a URL.
    """

    def __init__(
        self,
        media_type: str | None = None,
    ):
        if type(self) is __class__:
            raise Exception(
                "Cannot instantiate Asset directly; use `Asset.new()` instead"
            )

        # The MIME type of the asset
        self.media_type = media_type

    @overload
    @classmethod
    def new(cls, data: bytes, media_type: str | None = None) -> BytesAsset: ...

    @overload
    @classmethod
    def new(cls, data: Path, media_type: str | None = None) -> PathAsset: ...

    @overload
    @classmethod
    def new(cls, data: URL, media_type: str | None = None) -> UrlAsset: ...

    @classmethod
    def new(
        cls,
        data: bytes | Path | URL,
        media_type: str | None = None,
    ) -> Asset:
        key = (data, media_type)
        try:
            return _ASSETS[key]
        except KeyError:
            pass

        if isinstance(data, Path):
            asset = PathAsset(data, media_type)
        elif isinstance(data, URL):
            asset = UrlAsset(data, media_type)
        elif isinstance(data, (bytes, bytearray)):
            asset = BytesAsset(data, media_type)
        elif isinstance(data, str):
            raise TypeError(
                f"Cannot create asset from input {data!r}. Perhaps you meant to"
                f" pass a `pathlib.Path` or `rio.URL`?"
            )
        else:
            raise TypeError(f"Cannot create asset from input {data!r}")

        _ASSETS[key] = asset
        return asset

    @classmethod
    def from_image(
        cls,
        image: ImageLike,
        media_type: str | None = None,
    ) -> Asset:
        if isinstance(image, Image):
            file = io.BytesIO()
            image.save(file, format="PNG")
            image = file.getvalue()
            media_type = "image/png"
        elif media_type is None:
            # For some image formats, browsers are too stupid to display the
            # image correctly if we don't explicitly tell them the mime type.
            # Check for those formats.
            media_type = detect_important_image_types(image)

        return Asset.new(image, media_type)

    @abc.abstractmethod
    async def try_fetch_as_blob(self) -> tuple[bytes, str | None]:
        """
        Try to fetch the image as blob & media type. Raises a `ValueError` if
        fetching fails.
        """
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Asset):
            return NotImplemented

        if self.media_type != other.media_type:
            return False

        if type(self) != type(other):
            return False

        return self._eq(other)

    @abc.abstractmethod
    def _eq(self, other: Asset) -> bool:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def url(self) -> URL:
        """
        Returns the URL at which the asset can be accessed. The URL may or may
        not be relative.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _serialize(self, sess: rio.Session) -> str:
        """
        Make sure this asset is available under its URL, and return that URL,
        unescaped.
        """
        raise NotImplementedError


class HostedAsset(Asset):
    """
    Base class for assets that are hosted locally.
    """

    @property
    def secret_id(self) -> str:
        # The asset's id both uniquely identifies the asset, and is used as part
        # of the asset's URL.
        #
        # It is derived from the data, so that if the same file is to be hosted
        # multiple times only one instance is actually stored. Furthermore, this
        # allows the client to cache the asset efficiently, since the URL is
        # always the same.
        try:
            return self._secret_id
        except AttributeError:
            pass

        self._secret_id = self._get_secret_id()
        return self._secret_id

    @abc.abstractmethod
    def _get_secret_id(self) -> str:
        raise NotImplementedError

    @property
    def url(self) -> URL:
        return URL(f"/rio/asset/temp-{self.secret_id}")

    def _serialize(self, sess: rio.Session) -> str:
        sess._app_server.weakly_host_asset(self)
        return f"/rio/asset/temp-{self.secret_id}"


class BytesAsset(HostedAsset):
    def __init__(
        self,
        data: bytes | bytearray,
        media_type: str | None = None,
    ):
        super().__init__(media_type)

        self.data = data

    def _eq(self, other: BytesAsset) -> bool:
        return self.data == other.data

    async def try_fetch_as_blob(self) -> tuple[bytes, str | None]:
        return self.data, self.media_type

    def _get_secret_id(self) -> str:
        # TODO: Consider only hashing part of the data + media type + size
        # rather than processing everything
        return "b-" + _securely_hash_bytes_changes_between_runs(self.data).hex()


class PathAsset(HostedAsset):
    def __init__(
        self,
        path: os.PathLike | str,
        media_type: str | None = None,
    ):
        super().__init__(media_type)

        # Note: The path doesn't have to point to an existing file. Files can be
        # created and deleted at any time, so we have to handle that situation
        # gracefully anyway.
        self.path = Path(path)

    def _eq(self, other: PathAsset) -> bool:
        return self.path == other.path

    async def try_fetch_as_blob(self) -> tuple[bytes, str | None]:
        try:
            return self.path.read_bytes(), self.media_type
        except IOError:
            raise ValueError(f"Could not load asset from {self.path}")

    def _get_secret_id(self) -> str:
        return (
            "f-"
            + _securely_hash_bytes_changes_between_runs(
                str(self.path).encode("utf-8")
            ).hex()
        )


class UrlAsset(Asset):
    def __init__(
        self,
        url: URL,
        media_type: str | None = None,
    ):
        super().__init__(media_type)

        self._url = url

    def _eq(self, other: UrlAsset) -> bool:
        return self.url == other.url

    async def try_fetch_as_blob(self) -> tuple[bytes, str | None]:
        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(str(self._url))

                content_type = response.headers.get("Content-Type")
                if content_type is None:
                    content_type = "application/octet-stream"
                else:
                    content_type, _, _ = content_type.partition(";")

                return response.read(), content_type
        except httpx.HTTPError:
            raise ValueError(f"Could not fetch asset from {self._url}")

    @property
    def url(self) -> URL:
        return self._url

    def _serialize(self, sess: rio.Session) -> str:
        return self._url.human_repr()


def detect_important_image_types(image: ImageLike) -> str | None:
    if isinstance(image, Path):
        if image.suffix == ".svg":
            return "image/svg+xml"

    if isinstance(image, bytes):
        if b"<svg" in image:
            return "image/svg+xml"

    return None



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Switch",
    "SwitchChangeEvent",
]


@dataclass
class SwitchChangeEvent:
    is_on: bool


class Switch(FundamentalComponent):
    """
    # Switch

    An input for `True` / `False` values.

    Switches allow the user to toggle between an "on" and an "off" state. They
    thus correspond to a Python `bool` value. Use them to allow the user to
    enable or disable certain features, or to select between two options.


    ## Attributes:

    `is_on`: Whether the switch is currently in the "on" state.

    `is_sensitive`: Whether the switch should respond to user input.

    `on_change`: Triggered when the user toggles the switch.


    ## Example:

    A minimal example of a `Switch` will be shown:

    ```python
    rio.Switch(is_on=True)
    ```

    You can also bind the value of the switch to a state variable. This will
    update the state variable whenever the switch value changes:

    ```python
    class MyComponent(rio.Component):
        is_on: bool = False

        def build(self) -> rio.Component:
            return rio.Switch(
                is_on=self.bind().is_on,
                on_change=lambda event: print(
                    f"Switch is now {'on' if event.is_on else 'off'}"
                ),
            )
    ```

    You can also listen to changes in the switch by providing an on_change
    callback:

    ```python
    class MyComponent(rio.Component):
        is_on: bool = False

        def on_change(self, event: rio.SwitchChangeEvent):
            self.is_on = event.is_on
            # You can do whatever you here like printing the state
            print(f"Switch is now {'on' if event.is_on else 'off'}")

        def build(self) -> rio.Component:
            return rio.Switch(
                is_on=self.is_on,
                on_change=self.on_change,
            )
    ```
    """

    is_on: bool = False
    _: KW_ONLY
    is_sensitive: bool = True
    on_change: rio.EventHandler[SwitchChangeEvent] = None

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"is_on"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

        if "is_on" in delta_state and not self.is_sensitive:
            raise AssertionError(
                f"Frontend tried to set `Switch.is_on` even though `is_sensitive` is `False`"
            )

    async def _call_event_handlers_for_delta_state(self, delta_state: JsonDoc) -> None:
        # Trigger on_change event
        try:
            new_value = delta_state["is_on"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, bool), new_value
            await self.call_event_handler(
                self.on_change,
                SwitchChangeEvent(new_value),
            )


Switch._unique_id = "Switch-builtin"



from __future__ import annotations

from typing import *  # type: ignore

from . import class_container

__all__ = [
    "Spacer",
]


class Spacer(class_container.ClassContainer):
    """
    # Spacer

    Adds empty space.

    Spacers are invisible components which add empty space between other
    components. While similar effects can often be achieved using margins and
    alignment, code with spacers can sometimes be easier to read.

    Note that unlike most components in Rio, `Spacer` does not have a `natural`
    size. Therefore it defaults to a width and height of `grow`, as that is how
    they're frequently used.

    ## Example:

    A minimal example of `Spacer` will be shown:

    ```python
    rio.Spacer(height=5)
    ```
    """

    def __init__(
        self,
        *,
        width: float | Literal["grow"] = "grow",
        height: float | Literal["grow"] = "grow",
        key: str | None = None,
    ):
        """
        Args:
            width: How much space the spacer should take up horizontally.
            height: How much space the spacer should take up vertically.
        """

        super().__init__(
            None,
            ["rio-spacer"],
            key=key,
            width=width,
            height=height,
        )


# Make sure the component is recognized as `ClassContainer`, rather than a new
# component.
Spacer._unique_id = class_container.ClassContainer._unique_id



from __future__ import annotations

from typing import Literal

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Tooltip",
]


class Tooltip(FundamentalComponent):
    """
    # Tooltip

    A small pop-up window that appears when the user hovers over an element.

    `Tooltip` is a small pop-up window that appears when the user hovers over an
    element. It is commonly used to provide additional information about the
    element, such as a description or a hint.


    ## Attributes:

    `anchor`: The component to which the tooltip is anchored.

    `tip`: The text or component to display in the tooltip.

    `position`: The position of the tooltip relative to the anchor. It can be
        one of the following values: `left`, `top`, `right`, `bottom`.


    ## Example:

    A minimal example of a `Tooltip` will be shown:

    ```python
    rio.Tooltip(
        anchor=rio.Icon("material/info"),
        tip="This is a tooltip.",
        position="top",
    )
    ```
    """

    anchor: rio.Component
    tip_text: str | None
    tip_component: rio.Component | None
    position: Literal["left", "top", "right", "bottom"]

    # Impute a Text instance if a string is passed in as the tip
    def __init__(
        self,
        anchor: rio.Component,
        tip: str | rio.Component,
        position: Literal["left", "top", "right", "bottom"],
        *,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.anchor = anchor

        if isinstance(tip, str):
            self.tip_text = tip
            self.tip_component = None
        else:
            self.tip_text = None
            self.tip_component = tip

        self.position = position

        self._properties_set_by_creator_.update(("tip_text", "tip_component"))

    def __post_init__(self):
        if isinstance(self.tip_text, str):
            self.tip_component = rio.Text(self.tip_text)


Tooltip._unique_id = "Tooltip-builtin"



from __future__ import annotations

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Switcher",
]


class Switcher(FundamentalComponent):
    """
    # Switcher

    A container which can switch between different components.

    A `Switcher` is a container which can switch between different components.
    It is commonly used to switch between different views or modes. The
    `content` attribute can be used to change the currently displayed component.

    ## Attributes:

    `content`: The currently displayed component.


    ## Example:

    A minimal example of a `Switcher` will be shown:
    TODO: check if this is the correct example

    ```python
    rio.Switcher(content=rio.Text("Hello, world!"))
    ```
    """

    content: rio.Component | None


Switcher._unique_id = "Switcher-builtin"



import rio


# <component>
class SampleComponent(rio.Component):
    """
    This is a sample component. Components are the building blocks of Rio apps.

    Tip: An easy way to add components to your project is using the `rio`
    command:

    ```bash
    rio add component
    ```
    """

    pressed: bool = False

    def on_press(self) -> None:
        self.pressed = True

    def build(self) -> rio.Component:
        return rio.Column(
            rio.Text(
                "Pressed!" if self.pressed else "This is a sample component",
            ),
            rio.Button(
                "Press me!",
                on_press=self.on_press,
            ),
            spacing=1,
            align_x=0.5,
            align_y=0.5,
        )


# </component>



from __future__ import annotations

from typing import *  # type: ignore

import rio

__all__ = [
    "currently_building_component",
]


# Before a component is built, this value is set to that component. This allows newly
# created components to infer their creator, as well as session.
#
# - `Component`: The component that is currently being built
# - `None`: The app's build method is currently being called
currently_building_component: rio.Component | None = None


# Same as `currently_building_component`, but holding that component's session.
#
# - `Session`: The session that owns the currently building component
# - `None`: No build is currently in progress
currently_building_session: rio.Session | None = None


# This counter is increased each time a component is built. It can thus be used to
# uniquely identify the build generation of every component.
build_generation: int = 0
